/*
 * Copyright (C) 2019-2024 iopsys Software Solutions AB
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation
 *
 *		Author: Amin Ben Ramdhane <amin.benramdhane@pivasoftware.com>
 *		Author: Mohd Husaam Mehdi <husaam.mehdi@iopsys.eu>
 */

#include "dns_sd.h"
#include <json-c/json.h>

/*************************************************************
* ENTRY METHOD
**************************************************************/
static int browseDNSSDServiceInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	struct dm_data curr_data = {0};
	struct srv_info current_srv_info_obj = {0};

	json_object *ubus_result_browse = NULL, *service_objects = NULL;
	int id = 0;
	char *inst = NULL;

	// call ubus
	dmubus_call("umdns", "browse", UBUS_ARGS{0}, 0, &ubus_result_browse);

	// return from here if result empty
	DM_ASSERT(ubus_result_browse, bbfdm_set_fault_message(dmctx, "ubus call umdns browse did not work"));

	json_object_object_get_ex(ubus_result_browse, "discovered_services", &service_objects);
	json_object_object_foreach(service_objects, key, val) {
		char *proto_position = NULL;
		char *str = NULL;

		// example of key: "_http._tcp"
		current_srv_info_obj.application_protocol = dmstrdup(key);

		// for readability
		str = current_srv_info_obj.application_protocol;

		// find the point where transport protocol starts
		if (DM_STRSTR(str, "._tcp")) {
			proto_position = DM_STRSTR(str, "._tcp");
			current_srv_info_obj.transport_protocol_flag = TRANSPORT_TCP;
		} else if (DM_STRSTR(str, "._udp")) {
			proto_position = DM_STRSTR(str, "._udp");
			current_srv_info_obj.transport_protocol_flag = TRANSPORT_UDP;
		} else {
			continue;
		}

		// mark the end of application protocol
		*proto_position = '\0';

		// find and remove leading underscore
		proto_position = DM_STRSTR(str, "_");
		if (proto_position)
			current_srv_info_obj.application_protocol = ++proto_position;

		json_object_object_foreach(val, inner_key, inner_val) {
			// update our object
			current_srv_info_obj.instance_name = dmstrdup(inner_key);

			curr_data.json_object = inner_val;
			curr_data.additional_data = (void *)(&current_srv_info_obj);

			inst = handle_instance_without_section(dmctx, parent_node, ++id);

			if (DM_LINK_INST_OBJ(dmctx, parent_node, (void *)&curr_data, inst) == DM_STOP)
				break;
		}
	}

	return 0;
}

static int browseDNSSDServiceTextRecordInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	struct dm_data curr_data = {0};

	char *inst = NULL;
	char *key_value_str = NULL;
	char *dup_key_value_str = NULL;
	char *pch = NULL, *spch = NULL;

	key_value_str = dmjson_get_value_array_all(((struct dm_data *)prev_data)->json_object, ",", 1, "txt");

	if (key_value_str) {
		int id = 0;
		dup_key_value_str = dmstrdup(key_value_str);
		for (pch = strtok_r(dup_key_value_str, ",", &spch); pch != NULL; pch = strtok_r(NULL, ",", &spch)) {
			inst = handle_instance_without_section(dmctx, parent_node, ++id);

			curr_data.additional_data = (void *)pch;
			if (DM_LINK_INST_OBJ(dmctx, parent_node, (void *)&curr_data, inst) == DM_STOP)
				break;
		}
	}

	return 0;
}

/*************************************************************
* GET & SET PARAM
**************************************************************/
static int get_dns_sd_enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_option_value_fallback_def("umdns", "@umdns[0]", "enable", "1");
	return 0;
}

static int set_dns_sd_enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_boolean(ctx, value))
				return FAULT_9007;
			break;
		case VALUESET:
			dmuci_set_value("umdns", "@umdns[0]", "enable", value);
			break;
	}

	return 0;
}

static int get_dns_sd_service_number_of_entries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	int cnt = get_number_of_entries(ctx, data, instance, browseDNSSDServiceInst);
	dmasprintf(value, "%d", cnt);

	return 0;
}

static int get_dns_sd_service_instance_name(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *instance_name;

	instance_name = dmstrdup(((struct srv_info *)((struct dm_data *)data)->additional_data)->instance_name);
	if (DM_STRLEN(instance_name) > 64 ) {
		instance_name[64] = '\0';
	}

	*value = instance_name;
	return 0;
}

static int get_dns_sd_service_application_protocol(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *proto;

	proto = dmstrdup(((struct srv_info *)((struct dm_data *)data)->additional_data)->application_protocol);
	if (DM_STRLEN(proto) > 16 ) {
		proto[16] = '\0';
	}

	*value = proto;
	return 0;
}

static int get_dns_sd_service_transport_protocol(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	if (((struct srv_info *)((struct dm_data *)data)->additional_data)->transport_protocol_flag == TRANSPORT_TCP)
		*value = "TCP";
	else
		*value = "UDP";

	return 0;
}

static int get_dns_sd_service_domain(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *domain;

	domain = dmstrdup(dmjson_get_value(((struct dm_data *)data)->json_object, 1, "domain"));
	if (DM_STRLEN(domain) > 256 ) {
		domain[256] = '\0';
	}

	*value = domain;
	return 0;
}

static int get_dns_sd_service_port(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *port;

	port = dmstrdup(dmjson_get_value(((struct dm_data *)data)->json_object, 1, "port"));

	*value = port;
	return 0;
}

static int get_dns_sd_service_target(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *target;

	target = dmstrdup(dmjson_get_value(((struct dm_data *)data)->json_object, 1, "host"));

	*value = target;
	return 0;
}

static int get_dns_sd_service_status(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	// inactive services are not present in output of ubus
	*value = "LeaseActive";
	return 0;
}

static int get_dns_sd_service_last_update(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *last_updated;

	last_updated = dmstrdup(dmjson_get_value(((struct dm_data *)data)->json_object, 1, "last_update"));

	*value = last_updated;
	return 0;
}

static int get_dns_sd_service_host(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *ipv4_address = NULL;
	json_object *ipv4_array = NULL;
	unsigned idx = 0;
	char hosts_list[HOSTS_LIST_SIZE] = {0};

	dmjson_foreach_value_in_array(((struct dm_data *)data)->json_object, ipv4_array, ipv4_address, idx, 1, "ipv4") {
		bbfdm_get_references(ctx, MATCH_FIRST, "Device.Hosts.Host.", "IPAddress", ipv4_address, hosts_list, sizeof(hosts_list));
	}

	*value = dmstrdup(hosts_list);
	return 0;
}

static int get_dns_sd_service_time_to_live(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *ttl;

	ttl = dmstrdup(dmjson_get_value(((struct dm_data *)data)->json_object, 1, "ttl"));

	*value = ttl;
	return 0;
}

static int get_dns_sd_service_priority(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *priority;

	priority = dmstrdup(dmjson_get_value(((struct dm_data *)data)->json_object, 1, "priority"));

	*value = priority;
	return 0;
}

static int get_dns_sd_service_weight(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *weight;

	weight = dmstrdup(dmjson_get_value(((struct dm_data *)data)->json_object, 1, "weight"));

	*value = weight;
	return 0;
}

static int get_dns_sd_service_text_record_number_of_entries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	int cnt = get_number_of_entries(ctx, data, instance, browseDNSSDServiceTextRecordInst);
	dmasprintf(value, "%d", cnt);

	return 0;
}

static int get_dns_sd_service_text_record_key(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *key_value_str = dmstrdup((char *)((struct dm_data *)data)->additional_data);

	char *key_end_position = DM_STRSTR(key_value_str, "=");

	// mark end of key
	if (key_end_position)
		*key_end_position = '\0';

	*value = key_value_str;
	return 0;
}

static int get_dns_sd_service_text_record_value(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *key_value_str = dmstrdup((char *)((struct dm_data *)data)->additional_data);

	char *value_start_position = DM_STRSTR(key_value_str, "=");

	// point to start of value
	if (++value_start_position)
		key_value_str = value_start_position;

	*value = key_value_str;
	return 0;
}

/**********************************************************************************************************************************
*                                            OBJ & LEAF DEFINITION
***********************************************************************************************************************************/
/* *** Device.DNS.SD. *** */
DMOBJ tDNSPluginObj[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys, version*/
{"SD", &DMREAD, NULL, NULL, NULL, NULL, NULL, NULL, tDNSSDObj, tDNSSDParams, NULL, BBFDM_BOTH, NULL},
{0}
};

/* *** Device.DNS.SD. *** */
DMOBJ tDNSSDObj[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys, version*/
{"Service", &DMREAD, NULL, NULL, NULL, browseDNSSDServiceInst, NULL, NULL, tDNSSDServiceObj, tDNSSDServiceParams, NULL, BBFDM_BOTH, NULL},
{0}
};

DMLEAF tDNSSDParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type, version*/
{"Enable", &DMWRITE, DMT_BOOL, get_dns_sd_enable, set_dns_sd_enable, BBFDM_BOTH},
{"ServiceNumberOfEntries", &DMREAD, DMT_UNINT, get_dns_sd_service_number_of_entries, NULL, BBFDM_BOTH},
{0}
};

/* *** Device.DNS.SD.Service.{i}. *** */
DMOBJ tDNSSDServiceObj[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys, version*/
{"TextRecord", &DMREAD, NULL, NULL, NULL, browseDNSSDServiceTextRecordInst, NULL, NULL, NULL, tDNSSDServiceTextRecordParams, NULL, BBFDM_BOTH, NULL},
{0}
};

DMLEAF tDNSSDServiceParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type, version*/
{"InstanceName", &DMREAD, DMT_STRING, get_dns_sd_service_instance_name, NULL, BBFDM_BOTH},
{"ApplicationProtocol", &DMREAD, DMT_STRING, get_dns_sd_service_application_protocol, NULL, BBFDM_BOTH},
{"TransportProtocol", &DMREAD, DMT_STRING, get_dns_sd_service_transport_protocol, NULL, BBFDM_BOTH},
{"Domain", &DMREAD, DMT_STRING, get_dns_sd_service_domain, NULL, BBFDM_BOTH},
{"Port", &DMREAD, DMT_UNINT, get_dns_sd_service_port, NULL, BBFDM_BOTH},
{"Target", &DMREAD, DMT_STRING, get_dns_sd_service_target, NULL, BBFDM_BOTH},
{"Status", &DMREAD, DMT_STRING, get_dns_sd_service_status, NULL, BBFDM_BOTH},
{"LastUpdate", &DMREAD, DMT_TIME, get_dns_sd_service_last_update, NULL, BBFDM_BOTH},
{"Host", &DMREAD, DMT_STRING, get_dns_sd_service_host, NULL, BBFDM_BOTH, DM_FLAG_REFERENCE},
{"TimeToLive", &DMREAD, DMT_UNINT, get_dns_sd_service_time_to_live, NULL, BBFDM_BOTH},
{"Priority", &DMREAD, DMT_UNINT, get_dns_sd_service_priority, NULL, BBFDM_BOTH},
{"Weight", &DMREAD, DMT_UNINT, get_dns_sd_service_weight, NULL, BBFDM_BOTH},
{"TextRecordNumberOfEntries", &DMREAD, DMT_UNINT, get_dns_sd_service_text_record_number_of_entries, NULL, BBFDM_BOTH},
{0}
};

/* *** Device.DNS.SD.Service.{i}.TextRecord.{i}. *** */
DMLEAF tDNSSDServiceTextRecordParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type, version*/
{"Key", &DMREAD, DMT_STRING, get_dns_sd_service_text_record_key, NULL, BBFDM_BOTH},
{"Value", &DMREAD, DMT_STRING, get_dns_sd_service_text_record_value, NULL, BBFDM_BOTH},
{0}
};

DM_MAP_OBJ tDynamicObj[] = {
/* parentobj, nextobject, parameter */
{"Device.DNS.", tDNSPluginObj, NULL},
{0}
};
